from time import sleep, time

from mido import (
    Message,
    MidiFile,
    bpm2tempo,
    get_output_names,
    open_output,
    tick2second,
)
from mido.backends.rtmidi import Output


class MidiMessagePlayer:
    def __init__(
        self, midi_output_name: str, ticks_per_beat: int = 960, bpm: int = 120
    ):
        self.midi_output: Output = open_output(midi_output_name)
        self.ticks_per_beat = ticks_per_beat
        self.tempo = bpm2tempo(bpm)
        self.playback_start_time = time()
        self.input_time = 0.0

    def send(self, messages: list[Message]):
        for msg in messages:
            time_s = tick2second(
                msg.time, ticks_per_beat=self.ticks_per_beat, tempo=self.tempo
            )
            self.input_time += time_s
            current_playback_time = time() - self.playback_start_time
            duration_to_next_event_s = self.input_time - current_playback_time

            if duration_to_next_event_s > 0:
                sleep(duration_to_next_event_s)

            self.midi_output.send(msg)


def get_available_midi_output_names():
    return get_output_names()


def get_default_midi_output_name() -> str:
    output_names = get_available_midi_output_names()
    if len(output_names) == 0:
        raise Exception("No MIDI output devices found")
    return output_names[0]


def save_midi_file(midi: MidiFile, filename: str):
    midi.save(filename)



import click
import pkg_resources
from mido import Message

from henon2midi.ascii_art import AsciiArtCanvas
from henon2midi.base import create_midi_file_from_midi_generator
from henon2midi.henon_midi_generator import HenonMidiGenerator
from henon2midi.math import rescale_number_to_range
from henon2midi.midi import (
    MidiMessagePlayer,
    get_available_midi_output_names,
    get_default_midi_output_name,
)


@click.version_option()
@click.command()
@click.option(
    "-a",
    "--a-parameter",
    default=1.0,
    help="The a parameter for the Henon mapping.",
    show_default=True,
    type=float,
)
@click.option(
    "-i",
    "--iterations-per-orbit",
    default=100,
    help="The number of iterations per orbit.",
    show_default=True,
    type=int,
)
@click.option(
    "-m",
    "--midi-output-name",
    default=get_default_midi_output_name(),
    help=(
        "The name of the MIDI output device, "
        "Available: [{}]".format(", ".join(set(get_available_midi_output_names())))
    ),
    show_default=True,
    type=str,
)
@click.option(
    "--ticks-per-beat",
    default=960,
    help="The number of ticks per beat.",
    show_default=True,
    type=int,
)
@click.option(
    "--bpm",
    default=120,
    help="The beats per minute.",
    show_default=True,
    type=int,
)
@click.option(
    "--notes-per-beat",
    default=4,
    help="The number of notes per beat.",
    show_default=True,
    type=int,
)
@click.option(
    "--x-midi-parameter-mappings",
    default="note",
    help="The MIDI parameter mappings for the x data point.",
    show_default=True,
    type=str,
)
@click.option(
    "--y-midi-parameter-mappings",
    default="velocity,pan",
    help="The MIDI parameter mappings for the y data point.",
    show_default=True,
    type=str,
)
@click.option(
    "-r",
    "--starting-radius",
    default=0.0,
    help="The starting radius for the Henon mapping.",
    show_default=True,
    type=float,
)
@click.option(
    "-s",
    "--radial-step",
    default=0.01,
    help="The radial step for the Henon mapping.",
    show_default=True,
    type=float,
)
@click.option(
    "-o",
    "--out",
    default="henon.mid",
    help="The path to the output MIDI file.",
    show_default=True,
    type=str,
)
@click.option(
    "--draw-ascii-art",
    is_flag=True,
    help="Draw the Henon mapping in ASCII art.",
    type=bool,
)
@click.option("--sustain", is_flag=True, help="Turn the sustain on.", type=bool)
@click.option(
    "--clip",
    is_flag=True,
    help="Clip the MIDI messages to the range of the MIDI parameter.",
    type=bool,
)
def cli(
    a_parameter: float,
    iterations_per_orbit: int,
    midi_output_name: str,
    ticks_per_beat: int,
    bpm: int,
    notes_per_beat: int,
    x_midi_parameter_mappings: str,
    y_midi_parameter_mappings: str,
    starting_radius: float,
    radial_step: float,
    out: str,
    draw_ascii_art: bool,
    sustain: bool,
    clip: bool,
):
    """An application that generates midi from procedurally generated Henon mappings."""

    package = "henon2midi"
    version = pkg_resources.require(package)[0].version
    version_string = package + " v" + version + "\n\n"
    click.echo(version_string)

    midi_output_file_name = out
    if midi_output_name == "default":
        midi_output_name = get_default_midi_output_name()
    else:
        midi_output_name = midi_output_name
    ticks_per_beat = ticks_per_beat
    bpm = bpm
    notes_per_beat = notes_per_beat
    x_midi_parameter_mappings_set = set(x_midi_parameter_mappings.split(","))
    y_midi_parameter_mappings_set = set(y_midi_parameter_mappings.split(","))
    starting_radius = starting_radius
    radial_step = radial_step
    draw_ascii_art = draw_ascii_art
    sustain = sustain
    clip = clip

    options_string = (
        "Running with the following parameters. Use --help to see all available options.\n"
        f"\ta parameter: {a_parameter}\n"
        f"\titerations per orbit: {iterations_per_orbit}\n"
        f"\tmidi output name: {midi_output_name}\n"
        f"\tticks per beat: {ticks_per_beat}\n"
        f"\tbpm: {bpm}\n"
        f"\tnotes per beat: {notes_per_beat}\n"
        f"\tx midi parameter mappings: {x_midi_parameter_mappings_set}\n"
        f"\ty midi parameter mappings: {y_midi_parameter_mappings_set}\n"
        f"\tstarting radius: {starting_radius}\n"
        f"\tradial step: {radial_step}\n"
        f"\tout: {midi_output_file_name}\n"
        f"\tdraw ascii art: {draw_ascii_art}\n"
        f"\tsustain: {sustain}\n"
        f"\tclip: {clip}\n"
        f"\n"
    )
    click.echo(options_string)

    henon_midi_generator = HenonMidiGenerator(
        a_parameter=a_parameter,
        iterations_per_orbit=iterations_per_orbit,
        starting_radius=starting_radius,
        radial_step=radial_step,
        note_length_ticks=ticks_per_beat // notes_per_beat,
        sustain=sustain,
        clip=clip,
        x_midi_parameter_mappings=x_midi_parameter_mappings_set,
        y_midi_parameter_mappings=y_midi_parameter_mappings_set,
    )

    if midi_output_file_name:
        mid = create_midi_file_from_midi_generator(
            henon_midi_generator, ticks_per_beat=ticks_per_beat, bpm=bpm
        )
        mid.save(midi_output_file_name)

    if draw_ascii_art:
        ascii_art_canvas_width = 160
        ascii_art_canvas_height = 80
        ascii_art_canvas = AsciiArtCanvas(
            ascii_art_canvas_width, ascii_art_canvas_height
        )
    art_string = ""

    if midi_output_name:
        midi_message_player = MidiMessagePlayer(
            midi_output_name=midi_output_name, ticks_per_beat=ticks_per_beat, bpm=bpm
        )
        while True:
            messages = henon_midi_generator.next_midi_messages()
            current_iteration = henon_midi_generator.current_iteration
            current_orbit = (
                current_iteration // henon_midi_generator.iterations_per_orbit
            )
            try:
                midi_message_player.send(messages)
            except KeyboardInterrupt:
                midi_message_player.midi_output.reset()
                for note in range(128):
                    midi_message_player.midi_output.send(
                        Message("note_off", note=note, velocity=0)
                    )
                sustain_off_msg = Message("control_change", control=64, value=0)
                midi_message_player.midi_output.send(sustain_off_msg)
                midi_message_player.midi_output.close()
                exit()
            current_state_string = (
                f"Current iteration: {current_iteration}\n"
                f"Current orbit: {current_orbit + 1}\n"
                "\n"
            )

            if draw_ascii_art:
                art_string = get_ascii_art(
                    henon_midi_generator.current_data_point,
                    current_iteration,
                    henon_midi_generator.iterations_per_orbit,
                    ascii_art_canvas,
                )

            click.clear()
            screen_render = (
                version_string + options_string + current_state_string + art_string
            )
            click.echo(screen_render)


def get_ascii_art(
    data_point: tuple[float, float],
    current_iteration,
    iterations_per_orbit,
    ascii_art_canvas: AsciiArtCanvas,
) -> str:
    if current_iteration == 0:
        ascii_art_canvas.clear()
    x = data_point[0]
    y = data_point[1]
    draw_point_coord = (
        round(rescale_number_to_range(x, (-1.0, 1.0), (0, ascii_art_canvas.width - 1))),
        round(
            rescale_number_to_range(y, (-1.0, 1.0), (0, ascii_art_canvas.height - 1))
        ),
    )
    ascii_art_canvas.draw_point(draw_point_coord[0], draw_point_coord[1], ".")

    if current_iteration % iterations_per_orbit == 0:
        ascii_art_canvas.set_color("random")
    return ascii_art_canvas.generate_string()



import random


class AsciiArtCanvas:
    RESET_COLOR = "\033[0m"

    COLORS = {
        "black": "\033[0;30m",
        "red": "\033[0;31m",
        "green": "\033[0;32m",
        "yellow": "\033[0;33m",
        "blue": "\033[0;34m",
        "magenta": "\033[0;35m",
        "cyan": "\033[0;36m",
        "white": "\033[0;37m",
        "bright_black": "\033[1;30m",
        "bright_red": "\033[1;31m",
        "bright_green": "\033[1;32m",
        "bright_yellow": "\033[1;33m",
        "bright_blue": "\033[1;34m",
        "bright_magenta": "\033[1;35m",
        "bright_cyan": "\033[1;36m",
        "bright_white": "\033[1;37m",
    }

    def __init__(self, width: int = 120, height: int = 80):
        self.width = width
        self.height = height
        self.current_color = "white"
        self.canvas = [[" " for _ in range(width)] for _ in range(height)]

    def draw_point(self, x: int, y: int, character: str = "X"):
        color_escape_code = self.COLORS[self.current_color]
        self.canvas[y][x] = color_escape_code + character

    def set_color(self, color: str):
        if color == "random":
            color = random.choice(list(self.COLORS.keys()))
        elif color not in self.COLORS:
            raise Exception(f"Color {color} not supported")

        self.current_color = color

    def clear(self):
        self.canvas = [[" " for _ in range(self.width)] for _ in range(self.height)]

    def generate_string(self):
        return (
            self.RESET_COLOR
            + "\n".join(["".join(row) for row in self.canvas])
            + self.RESET_COLOR
        )



"""Tests for hello function."""
import pytest

from henon2midi.math import rescale_number_to_range


@pytest.mark.parametrize(
    ("value", "initial_range", "new_range", "expected"),
    [
        (-1.0, (-1.0, 1.0), (0, 127), 0.0),
        (0.0, (-1.0, 1.0), (0, 127), 63.5),
        (1.0, (-1.0, 1.0), (0, 127), 127.0),
        (-1.0, (-1.0, 1.0), (10, 20), 10.0),
        (0.0, (-1.0, 1.0), (10, 20), 15.0),
        (1.0, (-1.0, 1.0), (10, 20), 20.0),
    ],
)
def test_rescale_number_to_range(value, initial_range, new_range, expected):
    assert (
        rescale_number_to_range(value, initial_range, new_range, clip_value=False)
        == expected
    )


@pytest.mark.parametrize(
    ("value", "initial_range", "new_range", "expected"),
    [
        (-1000.0, (-1.0, 1.0), (0, 127), 0),
        (0.0, (-1.0, 1.0), (0, 127), 63.5),
        (100.0, (-1.0, 1.0), (0, 127), 127),
    ],
)
def test_rescale_number_to_range_with_clipping(
    value, initial_range, new_range, expected
):
    assert (
        rescale_number_to_range(value, initial_range, new_range, clip_value=True)
        == expected
    )


@pytest.mark.parametrize(
    ("value", "initial_range", "new_range"),
    [
        (-1000.0, (-1.0, 1.0), (0, 127)),
        (100.0, (-1.0, 1.0), (0, 127)),
    ],
)
def test_rescale_number_to_range_value_out_of_initial_range_raises_error(
    value, initial_range, new_range
):
    with pytest.raises(ValueError):
        rescale_number_to_range(value, initial_range, new_range, clip_value=False)



"""Tests for hello function."""
import pytest

from henon2midi.midi import get_default_midi_output_name


@pytest.fixture
def mock_get_output_names(mocker):
    mocker.patch("henon2midi.midi.get_output_names", return_value=["Bus 1", "Bus 2"])


@pytest.fixture
def mock_get_output_names_empty(mocker):
    mocker.patch("henon2midi.midi.get_output_names", return_value=[])


def test_get_default_midi_output_name(mock_get_output_names):
    assert get_default_midi_output_name() == "Bus 1"


def test_get_default_midi_output_name_empty(mock_get_output_names_empty):
    with pytest.raises(Exception):
        get_default_midi_output_name()



from math import cos, sin
from typing import Callable, Generator


def equation_a(x: float, y: float, a: float) -> float:
    return (x * cos(a)) - ((y - x**2) * sin(a))


def equation_b(x: float, y: float, a: float) -> float:
    return (x * sin(a)) + ((y - x**2) * cos(a))


# TODO: try these equations
# def four_parameter_equation_a(x: float, y: float, a: float, b: float, c: float, d: float) -> float:
#     return (sin(a*y)) - (cos(b*x))

# def four_parameter_equation_b(x: float, y: float, a: float, b: float, c: float, d: float) -> float:
#     return (sin(c*x)) - (cos(d*y))


def henon_mapping_generator(
    a_parameter: float,
    initial_x: float,
    initial_y: float,
    equation_a: Callable = equation_a,
    equation_b: Callable = equation_b,
) -> Generator[tuple[float, float], None, None]:
    x = initial_x
    y = initial_y
    while True:
        try:
            x_next, y_next = equation_a(x, y, a_parameter), equation_b(
                x, y, a_parameter
            )
            x, y = x_next, y_next
        except OverflowError:
            break
        else:
            yield x, y


def radially_expanding_henon_mappings_generator(
    a_parameter: float,
    iterations_per_orbit: int = 32,
    starting_radius: float = 0.0,
    radial_step: float = 0.1,
) -> Generator[tuple[float, float], None, None]:
    radius = starting_radius
    while radius <= 1:
        radius += radial_step
        for _ in range(iterations_per_orbit):
            yield from henon_mapping_generator(a_parameter, radius, radius)



def rescale_number_to_range(
    x: float,
    initial_range: tuple[float, float],
    new_range: tuple[float, float],
    clip_value: bool = True,
) -> float:
    if clip_value:
        if x < initial_range[0]:
            x = initial_range[0]
        elif x > initial_range[1]:
            x = initial_range[1]
    else:
        if x < initial_range[0] or x > initial_range[1]:
            raise ValueError(f"x ({x}) is not within initial_range ({initial_range})")

    initial_range_min = initial_range[0]
    initial_range_max = initial_range[1]
    initial_range_size = initial_range_max - initial_range_min

    new_range_min = new_range[0]
    new_range_max = new_range[1]
    new_range_size = new_range_max - new_range_min

    scale_factor = new_range_size / initial_range_size
    x_rescaled = ((x - initial_range_min) * scale_factor) + new_range_min

    return x_rescaled



from mido import MetaMessage, MidiFile, MidiTrack, bpm2tempo

from henon2midi.henon_midi_generator import HenonMidiGenerator


def create_midi_file_from_midi_generator(
    henon_midi_generator: HenonMidiGenerator,
    ticks_per_beat: int = 960,
    bpm: int = 120,
) -> MidiFile:
    mid = MidiFile(ticks_per_beat=ticks_per_beat)
    track = MidiTrack()
    mid.tracks.append(track)
    tempo = bpm2tempo(bpm)
    track.append(MetaMessage("set_tempo", tempo=tempo))
    track.extend(henon_midi_generator.generate_all_midi_messages())
    return mid



from mido import Message

from henon2midi.henon_equations import radially_expanding_henon_mappings_generator
from henon2midi.math import rescale_number_to_range


class HenonMidiGenerator:
    def __init__(
        self,
        a_parameter: float,
        iterations_per_orbit: int = 50,
        starting_radius: float = 0.0,
        radial_step: float = 0.05,
        note_length_ticks: int = 960,
        sustain: bool = False,
        clip: bool = False,
        x_midi_parameter_mappings: set[str] = {"note"},
        y_midi_parameter_mappings: set[str] = {"velocity", "pan"},
    ):
        self.a_parameter = a_parameter
        self.iterations_per_orbit = iterations_per_orbit
        self.starting_radius = starting_radius
        self.radial_step = radial_step
        self.note_length_ticks = note_length_ticks
        self.sustain = sustain
        self.clip = clip
        self.x_midi_parameter_mappings = x_midi_parameter_mappings
        self.y_midi_parameter_mappings = y_midi_parameter_mappings
        self.current_iteration = 0
        self.current_radius = self.starting_radius
        self.current_data_point = (self.starting_radius, self.starting_radius)
        self.current_iteration_midi_messages: list[Message] = []
        self.datapoint_generator = radially_expanding_henon_mappings_generator(
            a_parameter=self.a_parameter,
            iterations_per_orbit=self.iterations_per_orbit,
            starting_radius=self.starting_radius,
            radial_step=self.radial_step,
        )
        self.reset()

    def next_midi_messages(self) -> list[Message]:
        try:
            datapoint = next(self.datapoint_generator)
        except StopIteration:
            self.reset()
            midi_messages = self.current_iteration_midi_messages
        else:
            self.current_data_point = datapoint
            self.current_iteration += 1
            if self.current_iteration % self.iterations_per_orbit == 0:
                self.current_radius += self.radial_step
                self.datapoint_generator = radially_expanding_henon_mappings_generator(
                    a_parameter=self.a_parameter,
                    iterations_per_orbit=self.iterations_per_orbit,
                    starting_radius=self.current_radius,
                    radial_step=self.radial_step,
                )
            midi_messages = create_midi_messages_from_data_point(
                datapoint,
                duration_ticks=self.note_length_ticks,
                sustain=self.sustain,
                clip=self.clip,
                x_midi_parameter_mappings=self.x_midi_parameter_mappings,
                y_midi_parameter_mappings=self.y_midi_parameter_mappings,
            )
            self.current_iteration_midi_messages = midi_messages
        return midi_messages

    def reset(self):
        self.current_iteration = 0
        self.current_radius = self.starting_radius
        self.current_data_point = (self.starting_radius, self.starting_radius)
        self.current_iteration_midi_messages = create_midi_messages_from_data_point(
            self.current_data_point,
            duration_ticks=self.note_length_ticks,
            sustain=self.sustain,
            clip=self.clip,
            x_midi_parameter_mappings=self.x_midi_parameter_mappings,
            y_midi_parameter_mappings=self.y_midi_parameter_mappings,
        )
        self.datapoint_generator = radially_expanding_henon_mappings_generator(
            a_parameter=self.a_parameter,
            iterations_per_orbit=self.iterations_per_orbit,
            starting_radius=self.starting_radius,
            radial_step=self.radial_step,
        )

    def generate_all_midi_messages(self) -> list[Message]:
        self.reset()
        complete = False
        midi_messages = []
        midi_messages.extend(self.current_iteration_midi_messages)
        while not complete:
            midi_messages.extend(self.next_midi_messages())
            if self.current_iteration == 0:
                complete = True
        return midi_messages


def create_midi_messages_from_data_point(
    datapoint: tuple[float, float],
    duration_ticks: float = 960,
    sustain: bool = False,
    clip: bool = False,
    x_midi_parameter_mappings: set[str] = {"note"},
    y_midi_parameter_mappings: set[str] = {"velocity", "pan"},
) -> list[Message]:
    x = datapoint[0]
    y = datapoint[1]

    midi_values = {
        "note": 64,
        "velocity": 64,
        "pan": 64,
    }

    for x_midi_parameter_mapping in x_midi_parameter_mappings:
        if (x > 1.0 or x < -1.0) and not clip:
            midi_values[x_midi_parameter_mapping] = 0
        else:
            midi_values[x_midi_parameter_mapping] = midi_value_from_data_point_value(x)

    for y_midi_parameter_mapping in y_midi_parameter_mappings:
        if (y > 1.0 or y < -1.0) and not clip:
            midi_values[y_midi_parameter_mapping] = 0
        else:
            midi_values[y_midi_parameter_mapping] = midi_value_from_data_point_value(y)

    note_on = Message(
        "note_on",
        note=midi_values["note"],
        velocity=midi_values["velocity"],
    )
    note_off = Message(
        "note_off",
        note=midi_values["note"],
        velocity=midi_values["velocity"],
        time=duration_ticks,
    )

    note_messages = [note_on, note_off]
    pre_note_messages = []
    post_note_messages = []

    if "pan" in x_midi_parameter_mappings or "pan" in y_midi_parameter_mappings:
        pan = Message(
            "control_change",
            control=10,
            value=midi_values["pan"],
        )
        pre_note_messages.append(pan)

        reset_pan = Message(
            "control_change",
            control=10,
            value=64,
        )
        post_note_messages.append(reset_pan)

    if sustain:
        sustain_on_msg = Message(
            "control_change",
            control=64,
            value=127,
        )
        pre_note_messages.append(sustain_on_msg)
    else:
        sustain_off_msg = Message(
            "control_change",
            control=64,
            value=0,
        )
        post_note_messages.append(sustain_off_msg)

    messages = pre_note_messages + note_messages + post_note_messages

    return messages


def midi_value_from_data_point_value(
    x: float,
    data_point_range: tuple[float, float] = (-1.0, 1.0),
    midi_range: tuple[int, int] = (0, 127),
) -> int:
    min_midi_value = midi_range[0]
    max_midi_value = midi_range[1]

    min_data_point_value = data_point_range[0]
    max_data_point_value = data_point_range[1]

    return round(
        rescale_number_to_range(
            x,
            (min_data_point_value, max_data_point_value),
            (min_midi_value, max_midi_value),
            clip_value=True,
        )
    )
